#!/bin/bash

# Copyright (c) 2021, NVIDIA CORPORATION.  All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

# This is a script to do OTA update for no partition layout change
# Usage:
#    sudo ./nv_ota_update_without_layout_change.sh

# 1. Validate target board
# 2. Enable A/B redundancy if not enabled
# 3. Check whether ROOTFS A/B and unified A/B is enabled or not
# 4. Update bootloader if required
# 5. Update rootfs if required
#   5a. Update rootfs by running "nv_update_engine" if ROOTFS A/B is enabled
#   5b. Update rootfs by booting to recovery kernel if ROOTFS A/B is not enabled
set -e

BUP_IMAGE="bl_update_payload"
BUP_IMAGE_BL_ONLY="bl_only_payload"
BUP_IMAGE_KERNEL_ONLY="kernel_only_payload"
BUP_IMAGE_XUSB_ONLY="xusb_only_payload"
BUP_UPDATE_UTILITIES=( "nv_update_engine" "nv_bootloader_payload_updater" "nvbootctrl" )
CHIP_ID=
DEFAULT_ROOTFS="APP"
DEFAULT_ROOTFS_B="${DEFAULT_ROOTFS}_b"
ENTRY_TABLE_FILE="/opt/ota_package/entry_table"
ENTRY_TABLE_BACKUP="entry_table.ota_backup"
R32_IMAGES_DIR="images-R32-ToT"
ROOTFS_AB_ENABLED=0
ROOTFS_IMAGE="system.img"
UNIFIED_AB_ENABLED=0
UPDATE_BOOTLOADER=0
UPDATE_ROOTFS=0

get_update_control()
{
	# Determine whether to update bootloader only or update rootfs
	# only or update both bootloader and rootfs
	# Usage:
	#        get_update_control {work_dir} {_ret_update_bl} {_ret_update_rfs}
	local work_dir="${1}"
	local _ret_update_bl="${2}"
	local _ret_update_rfs="${3}"
	local update_control_file="${work_dir}/update_control"

	if [ ! -f "${update_control_file}" ]; then
		ota_log "The file ${update_control_file} is not found"
		return 1
	fi

	local line=
	local update_bl=0
	local update_rfs=0
	while read -r line
	do
		case ${line} in
		bootloader) update_bl=1; ;;
		rootfs) update_rfs=1; ;;
		*) break; ;;
		esac
	done <"${update_control_file}"

	if [ "${_ret_update_bl}" != "" ]; then
		eval "${_ret_update_bl}=${update_bl}"
	fi
	if [ "${_ret_update_rfs}" != "" ]; then
		eval "${_ret_update_rfs}=${update_rfs}"
	fi
	return 0
}

get_rootfs_a_b_enabled()
{
	# Check whether ROOTFS A/B is enabled and unified A/B is enabled.
	# Usage:
	#        get_rootfs_a_b_enabled {_ret_rootfs_ab_enabled} {_ret_unified_ab_enabled}
	local _ret_rootfs_ab_enabled="${1}"
	local _ret_unified_ab_enabled="${2}"
	local rootfs_ab_enabled=
	local unified_ab_enabled=
	local _nvbootctrl=
	_nvbootctrl="$(which nvbootctrl)"

	# Use the return code to differentiate rootfs AB and unified AB features:
	# EX_OK(0): Rootfs AB is enabled with unified AB.
	# EX_SOFTWARE(70): Rootfs AB is enabled without unified AB.
	# EX_UNAVAILABLE(69): Rootfs AB isn't enabled.
	local ret_code=
	"${_nvbootctrl}" is-unified-enabled >/dev/null 2>&1
	ret_code=$?
	if [ "${ret_code}" == 0 ]; then
		rootfs_ab_enabled=1
		unified_ab_enabled=1
	elif [ "${ret_code}" == 70 ]; then
		rootfs_ab_enabled=1
		unified_ab_enabled=0
	elif [ "${ret_code}" == 69 ]; then
		rootfs_ab_enabled=0
		unified_ab_enabled=0
	else
		ota_log "Failed to run \"${_nvbootctrl} is-unified-enabled\""
		return 1
	fi
	eval "${_ret_rootfs_ab_enabled}=${rootfs_ab_enabled}"
	eval "${_ret_unified_ab_enabled}=${unified_ab_enabled}"
	return 0
}

verify_a_b_status()
{
	# Verify A/B status by executing "nv_update_engine -v" command
	# Usage:
	#        verify_a_b_status
	local _nv_update_engine=
	_nv_update_engine="$(which nv_update_engine)"
	if ! "${_nv_update_engine}" -v; then
		ota_log "Failed to run \"${_nv_update_engine} -v\""
		return 1
	fi
	return 0
}

update_bootloader_with_UE()
{
	# Update bootloader on non-current slot by using "nv_update_engine" under
	# OTA work directory as BUP image has dependency on the the utility
	# "nv_update_engine" and the bootloader updater it invokes. Their versions
	# must be matched.
	# If update is successful, the entry table is backed up into ${work_dir}
	# Usage:
	#        update_bootloader_with_UE {work_dir} {chip_id}
	local work_dir="${1}"
	local chip_id="${2}"
	local bup_image="${work_dir}/${BUP_IMAGE}"
	local bl_only_image="${work_dir}/${BUP_IMAGE_BL_ONLY}"

	if [ ! -f "${bup_image}" ]; then
		ota_log "The BUP image ${bup_image} is not found"
		return 1
	fi
	if [ ! -f "${bl_only_image}" ]; then
		ota_log "The booloader only payload ${bl_only_image} is not found"
		return 1
	fi

	# Remove existing entry table before updating bootloader
	if [ -f "${ENTRY_TABLE_FILE}" ]; then
		ota_log "Remove exsting ${ENTRY_TABLE_FILE}"
	fi

	# Run "nv_update_engine -v" to make sure current status is clean
	if ! verify_a_b_status; then
		ota_log "Failed to run \"verify_a_b_status\""
		return 1
	fi

	# Run "nv_update_engine" to update bootloader
	# If ROOTFS A/B is enabled, the kernel/kernel-dtb partition is tied to rootfs,
	# so install the booloader only payload.
	# Otherwise, install the payload that contains both bootloader and
	# kernel/kernel-dtb.
	local ue_args=
	if [ "${ROOTFS_AB_ENABLED}" == 1 ]; then
		if [ "${UNIFIED_AB_ENABLED}" == 1 ]; then
			ue_args="-i bl-only --payload ${bl_only_image} --no-reboot"
		else
			ue_args="-i bl --payload ${bl_only_image} --no-reboot"
		fi
	else
		ue_args="-i bl --payload ${bup_image} --no-reboot"
	fi

	local _nv_update_engine=
	_nv_update_engine="$(which nv_update_engine)"
	if ! eval "${_nv_update_engine}" ${ue_args}; then
		ota_log "Failed to run \"${_nv_update_engine} ${ue_args}\""
		return 1
	fi

	# Run "nv_update_engine" to update xusb firmware for t19x devices
	local xusb_only_image="${work_dir}/${BUP_IMAGE_XUSB_ONLY}"
	if [ "${chip_id}" == "0x19" ]; then
		if [ "${UNIFIED_AB_ENABLED}" == 1 ]; then
			ue_args="-i bl-only --payload ${xusb_only_image} --no-reboot"
		else
			ue_args="-i bl --payload ${xusb_only_image} --no-reboot"
		fi
		if ! "${_nv_update_engine}" ${ue_args}; then
			ota_log "Failed to run \"${_nv_update_engine} ${ue_args}\""
			return 1
		fi
	fi

	# Back up the entry table that lists those partitions just updated.
	if [ ! -f "${ENTRY_TABLE_FILE}" ]; then
		ota_log "Entry table for BUP update is not found at ${ENTRY_TABLE_FILE}"
		return 1
	fi
	ota_log "Copy ${ENTRY_TABLE_FILE} into ${work_dir}/${ENTRY_TABLE_BACKUP}"
	cp "${ENTRY_TABLE_FILE}" "${work_dir}"/"${ENTRY_TABLE_BACKUP}"
	return 0
}

update_rootfs_with_UE()
{
	# Update rootfs on non-current slot by using "nv_update_engine"
	# It is only valid when ROOTFS A/B is enabled
	# Usage:
	#        update_rootfs_with_UE {work_dir} {update_rootfs_slot}
	local work_dir="${1}"
	local update_rootfs_slot="${2}"
	if [ "${ROOTFS_AB_ENABLED}" != 1 ]; then
		return 1
	fi

	local kernel_only_image="${work_dir}/${BUP_IMAGE_KERNEL_ONLY}"
	local rootfs_image="${work_dir}/${ROOTFS_IMAGE}"
	if [ ! -f "${kernel_only_image}" ]; then
		ota_log "The kernel only image ${kernel_only_image} is not found"
		return 1
	fi
	if [ ! -f "${rootfs_image}" ]; then
		ota_log "The rootfs image ${rootfs_image} is not found"
		return 1
	fi

	# Set rootfs updater
	local rootfs_updater=
	if ! set_rootfs_updater "${work_dir}" rootfs_updater; then
		ota_log "Failed to run \"set_rootfs_updater ${work_dir} rootfs_updater\""
		return 1
	fi

	# Run "nv_update_engine -v" to make sure current status is clean
	if ! verify_a_b_status; then
		ota_log "Failed to run \"verify_a_b_status\""
		return 1
	fi

	# Set rootfs name according to the "${update_rootfs_slot}"
	local rootfs_name=
	local suffix=""
	local rootfs_part_devnode=
	if [ "${update_rootfs_slot}" == "B" ]; then
		rootfs_name="${DEFAULT_ROOTFS_B}"
		suffix="_b"
	else
		rootfs_name="${DEFAULT_ROOTFS}"
	fi
	rootfs_part_devnode="${_PARTLABEL_PATH}/${rootfs_name}"
	if [ ! -e "${rootfs_part_devnode}" ];  then
		ota_log "Failed to find the ${rootfs_name} partition"
		return 1
	fi

	#
	# When ROOTFS A/B is enabled, the images update on kernel/kernel-dtb partitions
	# are tied to rootfs update. So we need to run "nv_bootloader_payload_updater"
	# with individual partition update option to update specific partition.
	#
	# When ROOTFS A/B is disabled, the images update on kernel/kernel-dtb partitions
	# are done along with bootloader update (in function update_bootloader_with_UE).
	#
	# Update kernel and kernel-dtb partition
	local _nv_bootloader_payload_updater=
	_nv_bootloader_payload_updater="$(which nv_bootloader_payload_updater)"
	cp -f "${kernel_only_image}" "/opt/ota_package/bl_update_payload"
	if ! eval "${_nv_bootloader_payload_updater}" --part "kernel${suffix}"; then
		ota_log "Failed to run \"${_nv_bootloader_payload_updater} --part kernel${suffix}\""
		return 1
	fi
	if ! eval "${_nv_bootloader_payload_updater}" --part "kernel-dtb${suffix}"; then
		ota_log "Failed to run \"${_nv_bootloader_payload_updater} --part kernel-dtb${suffix}\""
		return 1
	fi

	# Run "nv_update_engine" to update rootfs
	local ue_args=
	if [ "${UNIFIED_AB_ENABLED}" == 1 ]; then
		ue_args="-i rf-only "
	else
		ue_args="-i rf "
	fi
	ue_args+="--payload ${rootfs_image} --rf-updater \"${rootfs_updater} -p ${rootfs_part_devnode} -d ${work_dir}\" --no-reboot"

	local _nv_update_engine=
	_nv_update_engine="$(which nv_update_engine)"
	if ! eval "${_nv_update_engine}" ${ue_args}; then
		ota_log "Failed to run \"${_nv_update_engine} ${ue_args}\""
		return 1
	fi
	return 0
}

update_utilities_for_BUP_update()
{
	# Update the utilities that are necessary to update
	# bootloader with BUP. These utilities includes:
	# nv_update_engine, nvbootctrl and nv_bootloader_payload_updater.
	# Copy these utilities from {work_dir} into /usr/sbin
	# Usage:
	#        update_utilities_for_BUP_update {work_dir}
	local work_dir="${1}"
	local src=
	local dst=
	local item=
	for item in "${BUP_UPDATE_UTILITIES[@]}"
	do
		dst="/usr/sbin/${item}"
		src="${work_dir}/${item}"
		if [ ! -x "${src}" ]; then
			chmod 0755 "${src}"
		fi
		cp -f "${src}" "${dst}"
	done
	return 0
}

update_bootloader()
{
	# Update bootloader partitions
	# Usage:
	#        update_bootloader {work_dir} ${unified_ab_enabled} {chip_id}
	local work_dir="${1}"
	local unified_ab_enabled="${2}"
	local chip_id="${3}"
	local update_slot=

	# Get the non-current slot of bootloader to be updated
	ota_log "get_update_slot bootloader ${unified_ab_enabled} update_slot"
	if ! get_update_slot "bootloader" "${unified_ab_enabled}" "update_slot"; then
		ota_log "Failed to run \"get_update_slot bootloader ${unified_ab_enabled} update_slot\""
		return 1
	fi
	ota_log "update_slot=${update_slot}"

	# Update bootloader by running "nv_update_engine" with the BUP file
	# in the OTA payload package
	ota_log "update_bootloader_with_UE ${work_dir} ${chip_id}"
	if ! update_bootloader_with_UE "${work_dir}" "${chip_id}"; then
		ota_log "Failed to run \"update_bootloader_with_UE ${work_dir} ${chip_id}\""
		return 1
	fi
	ota_log "Bootloader on non-current slot(${update_slot}) is updated"
}

update_rootfs_with_a_b_enabled()
{
	# Update rootfs partition when ROOTFS A/B is enabled.
	# Usage:
	#        update_rootfs_with_a_b_enabled {work_dir} {unified_ab_enabled}
	local work_dir="${1}"
	local unified_ab_enabled="${2}"
	local update_rootfs_slot=

	# Get the rootfs slot to be updated
	ota_log "get_update_slot rootfs ${unified_ab_enabled} update_rootfs_slot"
	if ! get_update_slot "rootfs" "${unified_ab_enabled}" "update_rootfs_slot"; then
		ota_log "Failed to run \"get_update_slot rootfs ${unified_ab_enabled} update_rootfs_slot\""
		return 1
	fi
	ota_log "update_rootfs_slot=${update_rootfs_slot}"

	# Run "nv_update_engine" to update rootfs on non-current slot
	ota_log "update_rootfs_with_UE ${work_dir} ${update_rootfs_slot}"
	if ! update_rootfs_with_UE "${work_dir}" "${update_rootfs_slot}"; then
		ota_log "Failed to run \"update_rootfs_with_UE ${work_dir} ${update_rootfs_slot}\""
		return 1
	fi
	ota_log "Rootfs on non-current slot(${update_rootfs_slot}) is updated"

	# Write recovery and recovery-dtb partitions with ToT recovery image/dtb
	echo "update_recovery ${work_dir}"
	if ! update_recovery "${work_dir}"; then
		echo "Failed to run \"update_recovery ${work_dir}\""
		return 1
	fi

	# Delete OTA work directory
	rm -Rf "${work_dir}"
	ota_log "OTA update is completed, device will boot to the slot that is updated"
}

update_rootfs_with_a_b_disabled()
{
	# Update rootfs partition when ROOTFS A/B is disabled
	# Usage:
	#        update_rootfs_with_a_b_disabled {work_dir} {unified_ab_enabled}
	local work_dir="${1}"

	# Write recovery and recovery-dtb partition with base recovery image/dtb
	ota_log "write_base_recovery ${work_dir}"
	if ! write_base_recovery "${work_dir}"; then
		ota_log "Failed to run \"write_base_recovery ${work_dir}\""
		return 1
	fi

	# Set bootctrl to boot to recovery kernel on next boot
	ota_log "write_kernel_bootctrl ${work_dir} ${R32_IMAGES_DIR}"
	if ! write_kernel_bootctrl "${work_dir}" "${R32_IMAGES_DIR}"; then
		ota_log "Failed to run \"write_kernel_bootctrl ${work_dir} ${R32_IMAGES_DIR}\""
		return 1
	fi

	# For cboot that may try to load kernel from filesystem through extlinux.conf,
	# we need to temporarily hide extlinux.conf (and restore it back after OTA)
	# so that cboot is forcefully to load the recovery kernel from kernel partition.
	ota_log "backup_extlinux_conf ${work_dir}"
	if ! backup_extlinux_conf "${work_dir}"; then
		ota_log "Failed to run \"backup_extlinux_conf ${work_dir}\""
		exit 1
	fi
	ota_log "Rootfs is to be updated in recovery kernel once device is rebooted."
}

update_rootfs()
{
	# Update rootfs partition when ROOTFS A/B is disabled
	# Usage:
	#        update_rootfs_with_a_b_disabled {work_dir} {unified_ab_enabled}
	local work_dir="${1}"
	local unified_ab_enabled="${2}"

	# If rootfs A/B is not enabled, boot to recovery kernel and then
	# update rootfs in recovery kernel.
	# If rootfs A/B is enabled, update rootfs on the non-current slot
	# directly by running "nv_update_engine" utility
	if [ "${ROOTFS_AB_ENABLED}" != 1 ]; then
		ota_log "update_rootfs_with_a_b_disabled ${work_dir}"
		if ! update_rootfs_with_a_b_disabled "${work_dir}"; then
			ota_log "Failed to run \"update_rootfs_with_a_b_disabled ${work_dir}\""
			return 1
		fi
	else
		ota_log "update_rootfs_with_a_b_enabled ${work_dir} ${unified_ab_enabled}"
		if ! update_rootfs_with_a_b_enabled "${work_dir}" "${unified_ab_enabled}"; then
			ota_log "Failed to run \"update_rootfs_with_a_b_enabled ${work_dir} ${unified_ab_enabled}\""
			return 1
		fi
	fi
}

# Make sure that this script is running in root privilege
USERID=$(id -u)
if [ "${USERID}" -ne 0 ]; then
       ota_log "Please run this program as root."
       exit 0
fi

ota_log "Command: ${BASH_SOURCE[0]}"

source ./nv_ota_check_version.sh
source ./nv_ota_common.func
source ./nv_ota_customer.conf

# Check whether current board matches with OTA package target board
ota_log "check_target_board ${OTA_WORK_DIR} TARGET_BOARD"
if ! check_target_board "${OTA_WORK_DIR}" "TARGET_BOARD"; then
	ota_log "Failed to run \"check_target_board ${OTA_WORK_DIR} TARGET_BOARD\""
	exit 1
fi

# Get the chip id
ota_log "get_chip_id CHIP_ID"
if ! get_chip_id "CHIP_ID"; then
	ota_log "Failed to run \"get_chip_id CHIP_ID\""
	exit 1
fi

# The utilities for updating bootloader and rootfs must be updated as well,
# these utilities are: "nv_update_engine", "nv_bootloader_payload_updater" and "nvbootctrl".
ota_log "update_utilities_for_BUP_update ${OTA_WORK_DIR}"
if ! update_utilities_for_BUP_update "${OTA_WORK_DIR}"; then
	ota_log "Failed to run \"update_utilities_for_BUP_update ${OTA_WORK_DIR}\""
	exit 1
fi

# Enable A/B redundancy if A/B has not been enabled yet.
ota_log "enable_a_b_redundancy"
if ! enable_a_b_redundancy; then
	ota_log "Failed to run \"enable_a_b_redundancy\""
	exit 1
fi

# Check whether ROOTFS A/B is enabled and unified A/B is enabled
ota_log "get_rootfs_a_b_enabled ROOTFS_AB_ENABLED UNIFIED_AB_ENABLED"
if ! get_rootfs_a_b_enabled "ROOTFS_AB_ENABLED" "UNIFIED_AB_ENABLED"; then
	ota_log "Failed to run \"get_rootfs_a_b_enabled ROOTFS_AB_ENABLED UNIFIED_AB_ENABLED\""
	exit 1
fi
ota_log "ROOTFS_AB_ENABLED=${ROOTFS_AB_ENABLED}"
ota_log "UNIFIED_AB_ENABLED=${UNIFIED_AB_ENABLED}"

# For ROOTFS A/B is not enabled, make sure the base version built
# in OTA package matches with version on target board.
if [ "${ROOTFS_AB_ENABLED}" != 1 ]; then
	ota_log "check_bsp_version ${OTA_WORK_DIR} BASE_VERSION"
	if ! check_bsp_version "${OTA_WORK_DIR}" "" "BASE_VERSION"; then
		ota_log "Failed to run \"check_bsp_version ${OTA_WORK_DIR} BASE_VERSION\""
		exit 1
	fi
fi

# Get the value of "update_control" file to determine whether
# to update bootloader or/and rootfs
ota_log "get_update_control ${OTA_WORK_DIR} UPDATE_BOOTLOADER UPDATE_ROOTFS"
if ! get_update_control "${OTA_WORK_DIR}" "UPDATE_BOOTLOADER" "UPDATE_ROOTFS"; then
	ota_log "Failed to run \"get_update_control ${OTA_WORK_DIR} UPDATE_BOOTLOADER UPDATE_ROOTFS\""
	exit 1
fi
ota_log "UPDATE_BOOTLOADER=${UPDATE_BOOTLOADER}, UPDATE_ROOTFS=${UPDATE_ROOTFS}"

# Update bootloader on the non-current slot if UPDATE_BOOTLOADER is 1
if [ "${UPDATE_BOOTLOADER}" == 1 ]; then
	ota_log "update_bootloader ${OTA_WORK_DIR} ${UNIFIED_AB_ENABLED} ${CHIP_ID}"
	if ! update_bootloader "${OTA_WORK_DIR}" "${UNIFIED_AB_ENABLED}" "${CHIP_ID}"; then
		ota_log "Failed to run \"update_bootloader ${OTA_WORK_DIR} ${UNIFIED_AB_ENABLED} ${CHIP_ID}\""
		exit 1
	fi
fi

# Update rootfs if UPDATE_ROOTFS is 1
if [ "${UPDATE_ROOTFS}" == 1 ]; then
	ota_log "update_rootfs ${OTA_WORK_DIR} ${UNIFIED_AB_ENABLED}"
	if ! update_rootfs "${OTA_WORK_DIR}" "${UNIFIED_AB_ENABLED}"; then
		ota_log "Failed to run \"update_rootfs ${OTA_WORK_DIR} ${UNIFIED_AB_ENABLED}\""
		exit 1
	fi
fi

